import json
import os

from choreo.trajectory import (
    DifferentialSample,
    DifferentialTrajectory,
    EventMarker,
    SwerveSample,
    SwerveTrajectory,
    load_event_marker,
)
from wpilib import getDeployDirectory

TRAJ_SCHEMA_VERSION = 1


def load_differential_trajectory_string(
    trajectory_json_string: str,
) -> DifferentialTrajectory:
    """Load a differential trajectory from a string.

    Parameter ``trajectory_json_string``:
        The JSON string.
    """
    data = json.loads(trajectory_json_string)
    name = data["name"]
    try:
        version = int(data["version"])
        if version != TRAJ_SCHEMA_VERSION:
            raise ValueError(
                f"{name}.traj: Wrong version {version}. Expected {TRAJ_SCHEMA_VERSION}"
            )
    except ValueError:
        raise ValueError(
            f"{name}.traj: Wrong version {data['version']}. Expected {TRAJ_SCHEMA_VERSION}"
        )
    samples = [
        DifferentialSample(
            float(sample["t"]),
            float(sample["x"]),
            float(sample["y"]),
            float(sample["heading"]),
            float(sample["vl"]),
            float(sample["vr"]),
            float(sample["omega"]),
            float(sample["al"]),
            float(sample["ar"]),
            [float(x) for x in sample["fl"]],
            [float(y) for y in sample["fr"]],
        )
        for sample in data["trajectory"]["samples"]
    ]
    splits = [int(split) for split in data["trajectory"]["splits"]]
    # Add 0 as the first split index
    if len(splits) == 0 or splits[0] != 0:
        splits.insert(0, 0)
    events = list(
        filter(
            lambda marker: marker is not None,
            [load_event_marker(event) for event in data["events"]],
        )
    )

    return DifferentialTrajectory(data["name"], samples, splits, events)


def load_differential_trajectory(trajectory_name: str) -> DifferentialTrajectory:
    """Load a differential trajectory from a file.

    Parameter ``trajectory_name``:
        The path name in Choreo, which matches the file name in the deploy
        directory. Do not include ".traj" here.
    """
    with open(
        os.path.join(getDeployDirectory(), "choreo", trajectory_name + ".traj"),
        "r",
        encoding="utf-8",
    ) as trajectory_file:
        data = trajectory_file.read()
    return load_differential_trajectory_string(data)


def load_swerve_trajectory_string(trajectory_json_string: str) -> SwerveTrajectory:
    """Load a swerve trajectory from a string.

    Parameter ``trajectory_json_string``:
        The JSON string.
    """
    data = json.loads(trajectory_json_string)
    name = data["name"]
    try:
        version = int(data["version"])
        if version != TRAJ_SCHEMA_VERSION:
            raise ValueError(
                f"{name}.traj: Wrong version {version}. Expected {TRAJ_SCHEMA_VERSION}"
            )
    except ValueError:
        raise ValueError(
            f"{name}.traj: Wrong version {data['version']}. Expected {TRAJ_SCHEMA_VERSION}"
        )
    samples = [
        SwerveSample(
            float(sample["t"]),
            float(sample["x"]),
            float(sample["y"]),
            float(sample["heading"]),
            float(sample["vx"]),
            float(sample["vy"]),
            float(sample["omega"]),
            float(sample["ax"]),
            float(sample["ay"]),
            float(sample["alpha"]),
            [float(x) for x in sample["fx"]],
            [float(y) for y in sample["fy"]],
        )
        for sample in data["trajectory"]["samples"]
    ]
    splits = [int(split) for split in data["trajectory"]["splits"]]
    # Add 0 as the first split index
    if len(splits) == 0 or splits[0] != 0:
        splits.insert(0, 0)
    events = list(
        filter(
            lambda marker: marker is not None,
            [load_event_marker(event) for event in data["events"]],
        )
    )

    return SwerveTrajectory(data["name"], samples, splits, events)


def load_swerve_trajectory(trajectory_name: str) -> SwerveTrajectory:
    """Load a swerve trajectory from a file.

    Parameter ``trajectory_name``:
        The path name in Choreo, which matches the file name in the deploy
        directory. Do not include ".traj" here.
    """
    with open(
        os.path.join(getDeployDirectory(), "choreo", trajectory_name + ".traj"),
        "r",
        encoding="utf-8",
    ) as trajectory_file:
        data = trajectory_file.read()
    return load_swerve_trajectory_string(data)



from math import pi

from choreo.trajectory import DifferentialSample, SwerveSample
from choreo.util import field_dimensions

# Not accurate, but ChoreoLib doesn't actually keep old field dimensions
FIELD_LENGTH_2022 = field_dimensions.FIELD_LENGTH
FIELD_LENGTH_2024 = field_dimensions.FIELD_LENGTH
FIELD_WIDTH_2022 = field_dimensions.FIELD_WIDTH


def test_swerve_sample():
    sample = SwerveSample(
        0.0,
        1.0,
        2.0,
        3.0,
        4.0,
        5.0,
        6.0,
        7.0,
        8.0,
        9.0,
        [10.0, 11.0, 12.0, 13.0],
        [14.0, 15.0, 16.0, 17.0],
    )
    mirrored2024 = SwerveSample(
        0.0,
        FIELD_LENGTH_2024 - 1.0,
        2.0,
        pi - 3.0,
        -4.0,
        5.0,
        -6.0,
        -7.0,
        8.0,
        -9.0,
        [-11.0, -10.0, -13.0, -12.0],
        [15.0, 14.0, 17.0, 16.0],
    )
    rotated2022 = SwerveSample(
        0.0,
        FIELD_LENGTH_2022 - 1.0,
        FIELD_WIDTH_2022 - 2.0,
        pi + 3.0,
        -4.0,
        -5.0,
        6.0,
        -7.0,
        -8.0,
        9.0,
        [-10.0, -11.0, -12.0, -13.0],
        [-14.0, -15.0, -16.0, -17.0],
    )

    assert sample.flipped(2024) == mirrored2024
    assert sample.flipped(2022) == rotated2022


def test_zero_swerve_sample():
    sample = SwerveSample(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, [0, 0, 0, 0], [0, 0, 0, 0])
    mirrored2024 = SwerveSample(
        0,
        FIELD_LENGTH_2024 - 0,
        0,
        pi - 0,
        0,
        0,
        0,
        0,
        0,
        0,
        [-0.0, -0.0, -0.0, -0.0],
        [0, 0, 0, 0],
    )
    rotated2022 = SwerveSample(
        0,
        FIELD_LENGTH_2022,
        FIELD_WIDTH_2022,
        pi,
        0,
        0,
        0,
        0,
        0,
        0,
        [-0.0, -0.0, -0.0, -0.0],
        [-0.0, -0.0, -0.0, -0.0],
    )

    assert sample.flipped(2024) == mirrored2024
    assert sample.flipped(2022) == rotated2022


def test_differential_sample():
    sample = DifferentialSample(0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0)
    mirrored2024 = DifferentialSample(
        0.0, FIELD_LENGTH_2024 - 1.0, 2.0, pi - 3.0, 5.0, 4.0, -6.0, 8.0, 7.0, 10.0, 9.0
    )
    rotated2022 = DifferentialSample(
        0.0,
        FIELD_LENGTH_2022 - 1.0,
        FIELD_WIDTH_2022 - 2.0,
        pi + 3.0,
        4.0,
        5.0,
        6.0,
        7.0,
        8.0,
        9.0,
        10.0,
    )

    assert sample.flipped(2024) == mirrored2024
    assert sample.flipped(2022) == rotated2022


def test_zero_differential_sample():
    sample = DifferentialSample(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    mirrored2024 = DifferentialSample(
        0.0, FIELD_LENGTH_2024 - 0, 0.0, pi - 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
    )
    rotated2022 = DifferentialSample(
        0.0, FIELD_LENGTH_2022, FIELD_WIDTH_2022, pi, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
    )

    assert sample.flipped(2024) == mirrored2024
    assert sample.flipped(2022) == rotated2022



import os

import choreo

TRAJECTORY = """
{
 "name":"New Path",
 "version":1,
 "snapshot":{
  "waypoints":[
    {"x":0.0, "y":0.0, "heading":0.0, "intervals":9, "split":false, "fixTranslation":true, "fixHeading":true, "overrideIntervals":false},
    {"x":1.0, "y":0.0, "heading":0.0, "intervals":9, "split":false, "fixTranslation":false, "fixHeading":false, "overrideIntervals":false},
    {"x":2.0, "y":0.0, "heading":0.0, "intervals":9, "split":false, "fixTranslation":false, "fixHeading":false, "overrideIntervals":false},
    {"x":3.0, "y":0.0, "heading":0.0, "intervals":40, "split":false, "fixTranslation":true, "fixHeading":true, "overrideIntervals":false}],
  "constraints":[
    {"from":"first", "to":null, "data":{"type":"StopPoint", "props":{}}},
    {"from":"last", "to":null, "data":{"type":"StopPoint", "props":{}}},
    {"from":1, "to":2, "data":{"type":"PointAt", "props":{"x":1.5, "y":4.0, "tolerance":0.017453292519943295, "flip":false}}}]
 },
 "params":{
  "waypoints":[
    {"x":["0 m",0.0], "y":["0 m",0.0], "heading":["0 deg",0.0], "intervals":9, "split":false, "fixTranslation":true, "fixHeading":true, "overrideIntervals":false},
    {"x":["1 m",1.0], "y":["0 m",0.0], "heading":["0 deg",0.0], "intervals":9, "split":false, "fixTranslation":false, "fixHeading":false, "overrideIntervals":false},
    {"x":["2 m",2.0], "y":["0 m",0.0], "heading":["0 deg",0.0], "intervals":9, "split":false, "fixTranslation":false, "fixHeading":false, "overrideIntervals":false},
    {"x":["3 m",3.0], "y":["0 m",0.0], "heading":["0 rad",0.0], "intervals":40, "split":false, "fixTranslation":true, "fixHeading":true, "overrideIntervals":false}],
  "constraints":[
    {"from":"first", "to":null, "data":{"type":"StopPoint", "props":{}}},
    {"from":"last", "to":null, "data":{"type":"StopPoint", "props":{}}},
    {"from":1, "to":2, "data":{"type":"PointAt", "props":{"x":["1.5 m",1.5], "y":["4 m",4.0], "tolerance":["1 deg",0.017453292519943295], "flip":false}}}]
 },
 "trajectory":{
  "sampleType":"Swerve",
  "waypoints":[0.0,0.1,0.2,0.3],
  "samples":[
    {"t":0.0, "x":0.0, "y":0.0, "heading":0.0, "vx":0.0, "vy":0.0, "omega":0.0, "ax":0.0, "ay":0.0, "alpha":0.0, "fx":[0.0,0.0,0.0,0.0], "fy":[0.0,0.0,0.0,0.0]},
    {"t":1.0, "x":0.5, "y":0.1, "heading":0.2, "vx":3.0, "vy":3.0, "omega":10.0, "ax":20.0, "ay":20.0, "alpha":30.0, "fx":[100.0,200.0,300.0,400.0], "fy":[-100.0,-200.0,-300.0,-400.0]}
  ],
  "splits":[0],
  "forcesAvailable":false
 },
 "events":[
  {"name":"testEvent", "from":{"target":0, "targetTimestamp":0, "offset":{"exp":"0 s", "val":0.0}}, "event":null}
 ]
}
"""

CORRECT_SWERVE_TRAJECTORY = choreo.SwerveTrajectory(
    "New Path",
    [
        choreo.SwerveSample(
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ),
        choreo.SwerveSample(
            1.0,
            0.5,
            0.1,
            0.2,
            3.0,
            3.0,
            10.0,
            20.0,
            20.0,
            30.0,
            [100.0, 200.0, 300.0, 400.0],
            [-100.0, -200.0, -300.0, -400.0],
        ),
    ],
    [0],
    [choreo.EventMarker(0.0, "testEvent")],
)


def test_basic_parse():
    trajectory = choreo.load_swerve_trajectory(
        os.path.join(os.path.dirname(__file__), "resources", "swerve_test")
    )

    for i in range(0, 500):
        trajectory.sample_at(i * 0.01)


def test_deserialize_swerve_trajectory():
    trajectory = choreo.load_swerve_trajectory_string(TRAJECTORY)
    assert trajectory == CORRECT_SWERVE_TRAJECTORY


def test_event_markers():
    trajectory = choreo.load_swerve_trajectory(
        os.path.join(os.path.dirname(__file__), "resources", "swerve_test")
    )

    for event in trajectory.events:
        print(event.timestamp, event.event)
